# -*- coding: utf-8 -*-
"""
Created on Sat Mar 30 12:01:44 2019

@author: Magnus

The purpose of this script is to model and solve the famous "Farmers problem" using Pyomo. 
(As presented in Introduction to Stochastic Programming by Birge&Louveaux p. 4-5)

The farmer has a total of 500 acres of available land for planting. He can produce Wheat, Corn and Sugar Beets.
There are certain requirements to the production, and there is a wholesaler available that the farmer can trade with.
All available data are summarized below:
    
                        WHEAT       CORN        SUGAR BEETS
Yield(T/acre)           2.5         3           20
Planting cost ($/acre)  150         230         260
Selling price ($/T)     170         150         36 under 6000 T
                                                10 over 6000 T
Purchase price ($/T)    238         210         -
Min Requirement (T)     200         240         -

The farmer then introduces the following decision variables:
    x1 = acres devoted to wheat
    x2 = acres devoted to corn
    x3 = acres devoted to sugar beets
    w1 = tons of wheat sold
    y1 = tons of wheat purchased
    w2 = tons of corn sold
    y2 = tons of corn purchased
    w3 = tons of sugar beets sold at high price
    w4 = tons of sugar beets sold at low price
    
The mathematical formulation of the problem then becomes (aiming to minimize costs):
    
    min z = 150*x1 + 230*x2 + 260*x3 + 238*y1 + 210*y2 - 170*w1 - 150*w2 - 36*w3 - 10*w4
    s.t     x1 + x2 + x3 <= 500
            2.5*x1 + y1 - w1 >= 200
            3*x2 + y2 - w2 >= 240
            w3 + w4 <= 20*x3
            w3 <= 6000
            all variables >= 0
            
"""
#Initializing script and declearing packages
import numpy as np
import sys
import time
import pandas as pd
import pyomo.environ as pyo
from pyomo.opt import SolverFactory

def EasySetup():
    # In this setup, all data are put directly into pyomo model
    model = pyo.ConcreteModel()
    """
    Sets
    """
    model.Crops = pyo.Set(initialize=["W","C","B"]) # W=wheat, C=corn, B=beets

    """
    Variables
    """
    model.plant = pyo.Var(model.Crops, within=pyo.NonNegativeReals, initialize=0)
    model.sell = pyo.Var(model.Crops, within=pyo.NonNegativeReals, initialize=0)
    model.buy = pyo.Var(model.Crops, within=pyo.NonNegativeReals, initialize=0)
    model.sell_extra = pyo.Var(within=pyo.NonNegativeReals, initialize=0)
    
    """
    Objective
    """
    def Objective(model):
        cost_planting = 150*model.plant["W"] + 230*model.plant["C"] + 260*model.plant["B"]
        cost_buying = 238*model.buy["W"] + 210*model.buy["C"]
        income_sale = 170*model.sell["W"] + 150*model.sell["C"] + 36*model.sell["B"] + 10*model.sell_extra
        return(cost_planting + cost_buying - income_sale)
    model.OBJ =     pyo.Objective(rule = Objective, sense = pyo.minimize)
    
    """
    Constraints
    """
     # Total available land
    def CropYield(model) :
        return(sum(model.plant[p] for p in model.Crops) <= 500)
    model.CropYield     = pyo.Constraint(rule = CropYield)

    # Wheat requirement
    def WheatReq(model) :
        return(2.5*model.plant["W"] + model.buy["W"] - model.sell["W"] >= 200)
    model.WheatReq      = pyo.Constraint(rule=WheatReq)
    
    # Corn requirement
    def CornReq(model) :
        return(3*model.plant["C"] + model.buy["C"] - model.sell["C"] >= 240)
    model.CornReq       = pyo.Constraint(rule=CornReq)
    
    # Maximum total sale of beets
    def BeetSaleMax(model) :
        return(model.sell["B"] + model.sell_extra <= 20*model.plant["B"])
    model.BeetSaleMax       = pyo.Constraint(rule=BeetSaleMax)
    
    # Maximum sale of beets at high price
    def BeetHiMax(model) :
        return(model.sell["B"] <= 6000)
    model.BeetHiMax         = pyo.Constraint(rule=BeetHiMax)
    
    """
    Solving model
    """
     #Solver to use
    opt = SolverFactory("glpk")
        
    #Allow finding dual data
    model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
    
    #Solve the problem
    results = opt.solve(model, load_solutions = True)
        
    #Display results
    model.display()
    
    return()

def AdvancedSetup(): 
    """
    STRUCTURING INPUT DATA
    """
    Data = {} # Construnction a dictionary to hold data
    Data["Yield"] = {"wheat":2.5, "corn":3, "beets":20}
    Data["Cost"] = {"wheat":150, "corn":230, "beets":260}
    Data["Sell_Price"] = {"wheat":170, "corn":150, "beets":36}
    Data["Buy_Price"] = {"wheat":238, "corn":210}
    Data["Required"] = {"wheat":200, "corn":240}
    
    plants = ["wheat", "corn", "beets"] # Declaration of set containing the different plant types
    
    """
    CREATING THE MODEL
    """
    # Constructiong concrete model
    model = pyo.ConcreteModel()
    
    """
    Sets
    """
    # Set of the three differents plants
    model.plants = pyo.Set(initialize=plants, ordered=True)
    
    """
    Parameters
    """
    # Transfering data form dictionary to the pyomo model
    # The crop yield
    model.Yield = pyo.Param(model.plants, initialize=Data["Yield"])
    # The planting cost
    model.Cost = pyo.Param(model.plants, initialize=Data["Cost"])
    # The selling price at the wholesale
    model.SellPrice = pyo.Param(model.plants, initialize=Data["Sell_Price"])
    # The purchase price at wholesaler
    model.BuyPrice = pyo.Param(model.plants, initialize=Data["Buy_Price"])
    # Required production for the farmer's cattle feed
    model.Required = pyo.Param(model.plants, initialize=Data["Required"])
    # Price for the extra Sugar beets > 6000
    model.ExtraPrice = pyo.Param(initialize=10)
        
    """
    Variables
    """
    # Devoted acres of land
    model.acres = pyo.Var(model.plants, within=pyo.NonNegativeReals, initialize=0)
    # Tons sold
    model.sold = pyo.Var(model.plants, within=pyo.NonNegativeReals, initialize=0)
    # Tons purchased
    model.bought = pyo.Var(model.plants, within=pyo.NonNegativeReals, initialize=0)
    # Extra beets sold at lower price
    model.extra_beets = pyo.Var(within=pyo.NonNegativeReals, initialize=0)
    
    """
    Objective function
    """
    # Minimize total cost for the farmer
    def Objective(model) :
        planting_cost = sum(model.Cost[plant]*model.acres[plant] for plant in model.plants)
        buying_cost = sum(model.BuyPrice[p]*model.bought[p] for p in ["wheat","corn"])
        income_sales = sum(model.SellPrice[p]*model.sold[p] for p in model.plants) + model.ExtraPrice * model.extra_beets
        return(planting_cost + buying_cost - income_sales)
    model.OBJ   = pyo.Objective(rule = Objective, sense = pyo.minimize)
    
    """
    Constraints
    """
    # Total available land
    def CropYield(model) :
        return(sum(model.acres[p] for p in model.plants) <= 500)
    model.CropYield     = pyo.Constraint(rule = CropYield)
    
    # Required wheat
    def WheatReq(model) :
        return(model.Yield["wheat"]*model.acres["wheat"] + model.bought["wheat"] - model.sold["wheat"] >= 200)
    model.WheatReq      = pyo.Constraint(rule = WheatReq)
    
    # Required corn
    def CornReq(model) :
        return(model.Yield["corn"]*model.acres["corn"] + model.bought["corn"] - model.sold["corn"] >= 240)
    model.CornReq      = pyo.Constraint(rule = CornReq)
    
    # Maximum total beets sold
    def MaxBeet(model) :
        return(model.sold["beets"] + model.extra_beets <= model.Yield["beets"]*model.acres["beets"])
    model.MaxBeet       = pyo.Constraint(rule = MaxBeet)
    
    # Maximum beets sold at high price
    def MaxHiBeet(model) :
        return(model.sold["beets"] <= 6000)
    model.MaxHiBeet     = pyo.Constraint(rule = MaxHiBeet)
    
    """
    SOLVING THE MODEL
    """
    #Solver to use
    opt = SolverFactory("glpk")
        
    #Allow finding dual data
    model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
    
    #Solve the problem
    results = opt.solve(model, load_solutions = True)
        
    #Display results
    model.display()
    return()
    
# Running the desired model
#EasySetup()
AdvancedSetup()