# -*- coding: utf-8 -*-
"""
Created on Wed Apr  3 16:47:59 2019

@author: Magnus

This is an extention of the implemented farmers_problem, which includes stochastic behaviour of the crop yield. 
It is assumed that all crops experience the same type of year, ether good, bad or avarage. 
A good year refers to an increase of 20% of the mean yield, while a bad year is a 20% decrease.
It is assumed that the weather effects doesn't affect prices in the market.
"""
#Initializing script and declearing packages
import numpy as np
import sys
import time
import pandas as pd
import pyomo.environ as pyo
from pyomo.opt import SolverFactory

def AdvancedSetup(): 
    """
    STRUCTURING INPUT DATA
    """
    Data = {} # Construnction a dictionary to hold data
    Data["Yield"] = {"wheat":2.5, "corn":3, "beets":20}
    Data["Cost"] = {"wheat":150, "corn":230, "beets":260}
    Data["Sell_Price"] = {"wheat":170, "corn":150, "beets":36}
    Data["Buy_Price"] = {"wheat":238, "corn":210}
    Data["Required"] = {"wheat":200, "corn":240}
    
    times = ["t0", "t1"]
    scenarios = ["s0", "s1", "s2"]
    
    Data["Scen"] = {"t0":["s0", "s1"], "t1":scenarios} # For adding another scenario in time-step 1
    
    plants = ["wheat", "corn", "beets"] # Declaration of set containing the different plant types
    Data["Scenarios"] = {"s0": 1, "s1":1.2, "s2":0.8}
    Data["ScenAcre"] = {"s0": 1, "s1":0.5}
    scenariosacres = ["s0", "s1"]
    
    # Func for defining 2-dim Set
    def funct(a,b):
        return((d,c) for d in a for c in b[d])
        
    
    """
    CREATING THE MODEL
    """
    # Constructiong concrete model
    model = pyo.ConcreteModel()
    
    """
    Sets
    """
    # Set of the three differents plants
    model.plants = pyo.Set(initialize=plants, ordered=True)
    # Sets for scenarios and times
    model.scen = pyo.Set(initialize=scenarios, ordered=True)
    # Set for times
    model.time = pyo.Set(initialize=times, ordered=True)
    
    # Set for acres-scenarios
    model.scenacre = pyo.Set(initialize=scenariosacres, ordered=True)
    # Set containing all scenarios in all timesteps
    model.timescen = pyo.Set(dimen=2, initialize=funct(times,Data["Scen"]))
    
    """
    Parameters
    """
    # Transfering data form dictionary to the pyomo model (Change parameters that are different for each scenario)
    # The crop yield
    model.Yield = pyo.Param(model.plants, initialize=Data["Yield"])
    # The planting cost
    model.Cost = pyo.Param(model.plants, initialize=Data["Cost"])
    # The selling price at the wholesale
    model.SellPrice = pyo.Param(model.plants, initialize=Data["Sell_Price"])
    # The purchase price at wholesaler
    model.BuyPrice = pyo.Param(model.plants, initialize=Data["Buy_Price"])
    # Required production for the farmer's cattle feed
    model.Required = pyo.Param(model.plants, initialize=Data["Required"])
    # Price for the extra Sugar beets > 6000
    model.ExtraPrice = pyo.Param(initialize=10)
    # Scenario factor
    model.ScenFact = pyo.Param(model.scen, initialize=Data["Scenarios"])
    # Scenario acre factor
    model.ScenAcreFact = pyo.Param(model.scenacre, initialize=Data["ScenAcre"])
    """
    Variables
    """
    # Devoted acres of land (Which variables are time dependent)
    model.acres = pyo.Var(model.plants, model.scenacre, within=pyo.NonNegativeReals, initialize=0)
    # Tons sold
    model.sold = pyo.Var(model.plants, model.scen, model.scenacre, within=pyo.NonNegativeReals, initialize=0)
    # Tons purchased
    model.bought = pyo.Var(model.plants, model.scen, model.scenacre, within=pyo.NonNegativeReals, initialize=0)
    # Extra beets sold at lower price
    model.extra_beets = pyo.Var(model.scen, model.scenacre, within=pyo.NonNegativeReals, initialize=0)
    
    """
    Objective function
    """
    # Minimize total cost for the farmer
    def Objective(model) :
        planting_cost = sum(sum(model.Cost[plant]*model.acres[plant,s0] for plant in model.plants) for s0 in model.scenacre)
        buying_cost = sum(sum(sum(model.BuyPrice[p]*model.bought[p,s,s0] for p in ["wheat","corn"]) for s in model.scen) for s0 in model.scenacre)
        income_sales = sum(sum(sum(model.SellPrice[p]*model.sold[p,s,s0] for p in model.plants) + model.ExtraPrice * model.extra_beets[s,s0] for s in model.scen) for s0 in model.scenacre)
        return(0.5*planting_cost + 0.5*1/3 * buying_cost - 0.5*1/3 * income_sales)
    model.OBJ   = pyo.Objective(rule = Objective, sense = pyo.minimize)
    
    """
    Constraints
    """
    # Total available land
    def CropYield(model, s0) :
        return(sum(model.acres[p, s0] for p in model.plants) <= 500 * model.ScenAcreFact[s0])
    model.CropYield     = pyo.Constraint(model.scenacre, rule = CropYield)
    
    # Required wheat
    def WheatReq(model, s, s0) :
        return(model.Yield["wheat"]*model.ScenFact[s]*model.acres["wheat", s0] + model.bought["wheat",s , s0] - model.sold["wheat", s, s0] >= 200)
    model.WheatReq      = pyo.Constraint(model.scen, model.scenacre, rule = WheatReq)
    
    # Required corn
    def CornReq(model, s, s0) :
        return(model.Yield["corn"]*model.ScenFact[s]*model.acres["corn", s0] + model.bought["corn", s, s0] - model.sold["corn", s, s0] >= 240)
    model.CornReq      = pyo.Constraint(model.scen, model.scenacre, rule = CornReq)
    
    # Maximum total beets sold
    def MaxBeet(model, s, s0) :
        return(model.sold["beets", s, s0] + model.extra_beets[s, s0] <= model.Yield["beets"]*model.ScenFact[s]*model.acres["beets", s0])
    model.MaxBeet       = pyo.Constraint(model.scen, model.scenacre, rule = MaxBeet)
    
    # Maximum beets sold at high price
    def MaxHiBeet(model, s, s0) :
        return(model.sold["beets", s, s0] <= 6000)
    model.MaxHiBeet     = pyo.Constraint(model.scen, model.scenacre, rule = MaxHiBeet)
    
    """
    SOLVING THE MODEL
    """
    #Solver to use
    opt = SolverFactory("glpk")
        
    #Allow finding dual data
    model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
    
    #Solve the problem
    results = opt.solve(model, load_solutions = True)
        
    #Display results
    model.display()
    
    
    # Calculate objective value for each scenario:
    
    return(Data)
    
#Running the model    
Data = AdvancedSetup()